"use strict";

function _typeof(obj) { "@babel/helpers - typeof"; return _typeof = "function" == typeof Symbol && "symbol" == typeof Symbol.iterator ? function (obj) { return typeof obj; } : function (obj) { return obj && "function" == typeof Symbol && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }, _typeof(obj); }

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.DirectedGraph = void 0;

var _toollib = require("../common/toollib");

var _Stack = require("../Stack.js");

var _GraphBase2 = require("./GraphBase");

var _Queue = require("../Queue/Queue");

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); Object.defineProperty(Constructor, "prototype", { writable: false }); return Constructor; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); Object.defineProperty(subClass, "prototype", { writable: false }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = _getPrototypeOf(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = _getPrototypeOf(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return _possibleConstructorReturn(this, result); }; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } else if (call !== void 0) { throw new TypeError("Derived constructors may only return object or undefined"); } return _assertThisInitialized(self); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {})); return true; } catch (e) { return false; } }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf.bind() : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

/**
* @description
* 有向图
* 使用十字链表实现
* 
* @export
* @class DirectedGraph
*/
var DirectedGraph = /*#__PURE__*/function (_GraphBase) {
  _inherits(DirectedGraph, _GraphBase);

  var _super = _createSuper(DirectedGraph);

  function DirectedGraph() {
    var _this;

    var customizedComparer = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

    _classCallCheck(this, DirectedGraph);

    _this = _super.call(this, customizedComparer);

    _this.toString = function () {
      var str = '';

      this.__vertexes.forEach(function (v) {
        str += "".concat(v.data.toString(), " ->\r\n    Out: [");
        var arcOut = v.firstOut;

        while (arcOut) {
          str += "".concat(arcOut.end.data.toString(), ",");
          arcOut = arcOut.nextOut;
        }

        str += '],\r\n    In:   [';
        var arcIn = v.firstIn;

        while (arcIn) {
          str += "".concat(arcIn.start.data.toString()), arcIn = arcIn.nextIn;
        }

        str += ']\r\n';
      });

      return str;
    };

    return _this;
  }
  /**
   * @description 添加顶点
   *
   * @param {*} data 顶点数据
   * @returns 添加结果：成功true/失败false
   * @memberof DirectedGraph
   */


  _createClass(DirectedGraph, [{
    key: "addVertex",
    value: function addVertex(data) {
      if (data == null) throw new Error("parameter error: data is ".concat(data));
      if (this.hasVertex(data)) return false;
      var v = new VertexNode(data);

      this.__vertexes.push(v);

      return true;
    }
    /**
     * @description 添加/更新边
     *
     * @param {*} o1 顶点数据1
     * @param {*} o2 顶点数据2
     * @param {*} [info=null] 边的数据
     * @returns 添加/更新结果：成功true/失败fasle
     * @memberof DirectedGraph
     */

  }, {
    key: "setArc",
    value: function setArc(o1, o2) {
      var info = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

      var v1 = this.__findVertex(o1);

      var v2 = this.__findVertex(o2);

      var arc = null; // 检查该边是否存在

      if (v1 && v2) {
        arc = this.__findArc(v1, v2);
      } else {
        if (!v1) {
          v1 = new VertexNode(data1);

          this.__vertexes.push(v1);
        }

        if (!v2) {
          v2 = new VertexNode(data2);

          this.__vertexes.push(v2);
        }
      } // 不存在：添加边


      if (!arc) {
        arc = new Arc(v1, v2, info); // 从 v1 到 v2，画一条边线(v1 的出边，v2 的入边)

        this.__addOutArc(v1, arc);

        this.__addInArc(v2, arc);
      } // 存在：更新边（权重值等信息）
      else {
        arc.info = info;
      }

      return true;
    }
    /**
     * @description 广度优先遍历
     *
     * @param {*} start 遍历的起始顶点
     * @param {function} callback 回调函数
     * @memberof UndirectedGraph
     */

  }, {
    key: "bfsTraverse",
    value: function bfsTraverse(start, callback) {
      if (this.vertexCount === 0) return;
      var vertex = null; // 参数检查：
      // --1. 两个参数的时候：指定起点和遍历的回调函数

      if (start && callback) {
        if (!(0, _toollib.isFunction)(callback)) throw new Error("".concat(callback, " is not a function."));
        vertex = this.__findVertex(start);
        if (vertex === undefined) throw new Error("vertex of ".concat(start, " doesnot exist."));
      } // --2.1 一个参数的时候，仅指定遍历的回调函数
      else if ((0, _toollib.isFunction)(start) && !callback) {
        callback = start;
        start = null;
        vertex = this.__vertexes[0];
      } // --2.2 两个参数的时候，仅指定遍历的回调函数
      else if (start == null && (0, _toollib.isFunction)(callback)) {
        vertex = this.__vertexes[0];
      } // --3. 参数错误的其它情况
      else {
        throw new Error("parameter error: need [vStart, callback] or [callback].\r\nreceived: ".concat(start, ", ").concat(callback));
      } // 广度优先：使用队列


      var queue = new _Queue.Queue(); // 初始化探索状态

      var map = this.__initVisitStatus();

      while (vertex) {
        // 如果当前顶点未被探索，则探索（回调访问）
        if (map.get(vertex) !== _GraphBase2.VisitStatus.VISITED) {
          var stop = callback(vertex.data);
          if (stop) return;
          map.set(vertex, _GraphBase2.VisitStatus.VISITED);
        }

        var arc = vertex.firstOut;

        while (arc) {
          var vEnd = arc.end; // 如果另一端顶点未被发现或探索，则加入队列并标记

          if (map.get(vEnd) === _GraphBase2.VisitStatus.INIT) {
            queue.enqueue(vEnd);
            map.set(vEnd, _GraphBase2.VisitStatus.DISCOVERED);
          } // 下一组出边


          arc = arc.nextOut;
        } // 下一层顶点


        vertex = queue.dequeue() || this.__getNextVertex(map);
      }
    }
    /**
     * @description 深度优先遍历
     *
     * @param {*} start 遍历的起始顶点
     * @param {function} callback 回调函数
     * @memberof UndirectedGraph
     */

  }, {
    key: "dfsTraverse",
    value: function dfsTraverse(start, callback) {
      if (this.vertexCount === 0) return;
      var vertex = null; // 参数检查：
      // --1. 两个参数的时候：指定起点和遍历的回调函数

      if (start && callback) {
        if (!(0, _toollib.isFunction)(callback)) throw new Error("".concat(callback, " is not a function."));
        vertex = this.__findVertex(start);
        if (vertex === undefined) throw new Error("vertex of ".concat(start, " doesnot exist."));
      } // --2.1 一个参数的时候，仅指定遍历的回调函数
      else if ((0, _toollib.isFunction)(start) && !callback) {
        callback = start;
        start = null;
        vertex = this.__vertexes[0];
      } // --2.2 两个参数的时候，仅指定遍历的回调函数
      else if (start == null && (0, _toollib.isFunction)(callback)) {
        vertex = this.__vertexes[0];
      } // --3. 参数错误的其它情况
      else {
        throw new Error("parameter error: need [vStart, callback] or [callback].\r\nreceived: ".concat(start, ", ").concat(callback));
      } // 深度优先：使用栈


      var stack = new _Stack.Stack(); // 初始化探索状态

      var map = this.__initVisitStatus();

      while (vertex) {
        // 如果未探索该顶点，则探索（回调）并标记
        if (map.get(vertex) !== _GraphBase2.VisitStatus.VISITED) {
          var stop = callback(vertex.data);
          if (stop) return;
          map.set(vertex, _GraphBase2.VisitStatus.VISITED);
          stack.push(vertex);
        }

        var arc = vertex.firstOut;

        while (arc) {
          var vNext = arc.end; // 如果未探索，则探索（回调）并标记

          if (map.get(vNext) !== _GraphBase2.VisitStatus.VISITED) {
            callback(vNext.data);
            map.set(vNext, _GraphBase2.VisitStatus.VISITED); // 记录该顶点以备回溯

            stack.push(vNext);
            arc = vNext.firstOut;
          } // 回溯后探索其它分支
          else {
            arc = arc.nextOut;
          }
        } // 回溯或跳转顶点


        vertex = stack.pop() || this.__getNextVertex(map);
      }
    }
    /**
     * @description 给顶点添加入边
     * 
     * @param {VertexNode} v
     * @param {Arc} arc
     */

  }, {
    key: "__addInArc",
    value: function __addInArc(v, arc) {
      var adjArc = v.firstIn; // 没有入边

      if (adjArc === null) {
        v.firstIn = arc;
      } // 已有入边
      else {
        while (adjArc.nextIn) {
          adjArc = adjArc.nextIn;
        }

        adjArc.nextIn = arc;
      }
    }
    /**
     * @description 给顶点添加出边
     * 
     * @param {VertexNode} v
     * @param {Arc} arc
     */

  }, {
    key: "__addOutArc",
    value: function __addOutArc(v, arc) {
      var adjArc = v.firstOut; // 没有出边

      if (adjArc === null) {
        v.firstOut = arc;
      } // 已有出边
      else {
        while (adjArc.nextOut) {
          adjArc = adjArc.nextOut;
        }

        adjArc.nextOut = arc;
      }
    }
    /**
     * @description 查找边
     *
     * @param {VertexNode} v1 顶点1
     * @param {VertexNode} v2 顶点2
     * @returns Arc | undefined
     * @memberof DirectedGraph
     */

  }, {
    key: "__findArc",
    value: function __findArc(v1, v2) {
      var arc = v1.firstOut;

      while (arc) {
        if (v2 === arc.end) return arc;
        arc = arc.nextOut;
      }
    }
    /**
     * @description 
     * 如果指定的起始顶点没有出边，
     * 则无法进行后面其它顶点的遍历，
     * 这里找出下一个未访问过的顶点，
     * 继续循环遍历（广度/深度优先）
     *
     * @param {Map<VertexNode, VisitStatus>} map
     * @memberof DirectedGraph
     */

  }, {
    key: "__getNextVertex",
    value: function __getNextVertex(map) {
      var list = Array.from(map);
      var statusKV = list.find(function (kv) {
        return kv[1] !== _GraphBase2.VisitStatus.VISITED;
      });
      var vertex = statusKV && statusKV[0];
      return vertex;
    }
  }]);

  return DirectedGraph;
}(_GraphBase2.GraphBase);
/**
* @description 有向图内部用顶点类
*
* @class VertexNode
*/


exports.DirectedGraph = DirectedGraph;

var VertexNode = /*#__PURE__*/function () {
  function VertexNode(data) {
    _classCallCheck(this, VertexNode);

    if (data == null) throw new Error("vertex cannot be null");
    this.data = data;
    this.firstIn = null; // Arc

    this.firstOut = null; // Arc
  }

  _createClass(VertexNode, [{
    key: "toString",
    value: function toString() {
      return "".concat(this.data.toString());
    }
  }]);

  return VertexNode;
}();
/**
* @description 有向图内部用边类
*
* @class Arc
*/


var Arc = /*#__PURE__*/function () {
  function Arc(start, end) {
    var info = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

    _classCallCheck(this, Arc);

    this.start = start; // VertexNode

    this.end = end; // VertexNode

    this.nextIn = null; // Arc

    this.nextOut = null; // Arc

    this.info = info; // Weight info etc.
  }

  _createClass(Arc, [{
    key: "toString",
    value: function toString() {
      return "\n          {\n              start: ".concat(this.start.toString(), ",\n              end: ").concat(this.end.toString(), ",\n              info: ").concat(this.info ? this.info.toString() : null, ",\n          }\n      ");
    }
  }]);

  return Arc;
}();