"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.VisitStatus = exports.GraphBase = void 0;

var _toollib = require("../common/toollib.js");

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); Object.defineProperty(Constructor, "prototype", { writable: false }); return Constructor; }

/**
* @description 图的基类
*
* @export GraphBase
* @class GraphBase
*/
var GraphBase = /*#__PURE__*/function () {
  function GraphBase() {
    var customizedComparer = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

    _classCallCheck(this, GraphBase);

    this.__comparator = (0, _toollib.eqComparer)(customizedComparer);
    this.__vertexes = []; // this.__arcs = []
  }
  /**
   * 顶点个数
   *
   * @readonly
   * @memberof GraphBase
   */


  _createClass(GraphBase, [{
    key: "vertexCount",
    get: function get() {
      return this.__vertexes.length;
    }
    /**
     * 边的条数
     *
     * @readonly
     * @memberof GraphBase
     */
    // get arcCount() {
    //     return this.__arcs.length
    // }

    /**
     * @description 检查顶点是否存在
     *
     * @param {*} data 顶点数据
     * @returns 存在true/不存在false
     * @memberof DirectedGraph
     */

  }, {
    key: "hasVertex",
    value: function hasVertex(data) {
      if (data == null) throw new Error("parameter error: data is ".concat(data));

      var ret = this.__findVertex(data);

      return ret !== undefined ? true : false;
    }
    /**
     * @description 检查边是否存在
     *
     * @param {*} o1 顶点数据1
     * @param {*} o2 顶点数据2
     * @returns 存在true/不存在false
     * @memberof DirectedGraph
     */

  }, {
    key: "hasArc",
    value: function hasArc(o1, o2) {
      if (!o1 || !o2) return false;

      var v1 = this.__findVertex(o1);

      var v2 = this.__findVertex(o2);

      if (!v1 || !v2) return false;

      var arc = this.__findArc(v1, v2);

      return arc !== undefined ? true : false;
    }
    /**
     * @description 初始化所以顶点的探索状态
     *
     * @returns Map
     * @memberof UndirectedGraph
     */

  }, {
    key: "__initVisitStatus",
    value: function __initVisitStatus() {
      var map = new Map();

      this.__vertexes.forEach(function (v) {
        map.set(v, VisitStatus.INIT);
      });

      return map;
    }
    /**
     * @description 查找顶点
     *
     * @param {*} data 顶点数据
     * @returns 顶点/undefined
     * @memberof DirectedGraph
     */

  }, {
    key: "__findVertex",
    value: function __findVertex(data) {
      var _this = this;

      if (this.vertexCount === 0) return undefined;
      return this.__vertexes.find(function (v) {
        return _this.__comparator(v.data, data);
      });
    }
    /**
     * @abstract
     * @description 查找边
     * 
     * @param {VertexNode} v1 顶点1
     * @param {VertexNode} v2 顶点2
     * @returns Arc | undefined
     * @memberof DirectedGraph
     */

  }, {
    key: "__findArc",
    value: function __findArc(v1, v2) {
      throw new Error("must be implemented by child class");
    }
  }]);

  return GraphBase;
}();
/**
* @description 探索标记
*/


exports.GraphBase = GraphBase;
var VisitStatus = {
  INIT: 'Init',
  // 初始状态
  DISCOVERED: 'Discovered',
  // 已发现未探索
  VISITED: 'Visited' // 已探索

};
exports.VisitStatus = VisitStatus;