"use strict";

function _typeof(obj) { "@babel/helpers - typeof"; return _typeof = "function" == typeof Symbol && "symbol" == typeof Symbol.iterator ? function (obj) { return typeof obj; } : function (obj) { return obj && "function" == typeof Symbol && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }, _typeof(obj); }

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.UndirectedGraph = void 0;

var _GraphBase2 = require("./GraphBase");

var _toollib = require("../common/toollib");

var _Stack = require("../Stack");

var _Queue = require("../Queue/Queue.js");

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); Object.defineProperty(Constructor, "prototype", { writable: false }); return Constructor; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); Object.defineProperty(subClass, "prototype", { writable: false }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = _getPrototypeOf(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = _getPrototypeOf(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return _possibleConstructorReturn(this, result); }; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } else if (call !== void 0) { throw new TypeError("Derived constructors may only return object or undefined"); } return _assertThisInitialized(self); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {})); return true; } catch (e) { return false; } }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf.bind() : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

/**
* @description 
* 无向图
* 使用邻接多重表实现
* 
* @export UndirectedGraph
* @class UndirectedGraph
*/
var UndirectedGraph = /*#__PURE__*/function (_GraphBase) {
  _inherits(UndirectedGraph, _GraphBase);

  var _super = _createSuper(UndirectedGraph);

  function UndirectedGraph() {
    var _this;

    var customizedComparer = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

    _classCallCheck(this, UndirectedGraph);

    _this = _super.call(this, customizedComparer);
    /**
     * 用字符串表示出该图的信息
     * 显示所有的邻接点
     */

    _this.toString = function () {
      var str = '';

      this.__vertexes.forEach(function (v) {
        str += "".concat(v.data.toString(), " -> [");
        var arc = v.first;

        while (arc) {
          if (v === arc.vertex1) {
            str += "".concat(arc.vertex2.data.toString(), ",");
            arc = arc.orderNext;
          } else {
            // v === arc.vertex2
            str += "".concat(arc.vertex1.data.toString(), ",");
            arc = arc.reverseNext;
          }
        }

        str += ']\r\n';
      });

      return str;
    };

    return _this;
  }
  /**
   * @description 添加顶点
   *
   * @param {*} data
   * @memberof UndirectedGraph
   */


  _createClass(UndirectedGraph, [{
    key: "addVertex",
    value: function addVertex(data) {
      if (data == null) throw new Error("parameter error: data is ".concat(data));
      if (this.hasVertex(data)) return false;
      var v = new VertexNode(data);

      this.__vertexes.push(v);

      return true;
    }
    /**
     * @description 添加或更新邻接边
     *
     * @param {*} data1
     * @param {*} data2
     * @param {*} info
     * @memberof UndirectedGraph
     */

  }, {
    key: "setArc",
    value: function setArc(data1, data2, info) {
      var v1 = this.__findVertex(data1);

      var v2 = this.__findVertex(data2);

      var arc;

      if (v1 && v2) {
        arc = this.__findArc(v1, v2);
      } else {
        if (!v1) {
          v1 = new VertexNode(data1);

          this.__vertexes.push(v1);
        }

        if (!v2) {
          v2 = new VertexNode(data2);

          this.__vertexes.push(v2);
        }
      } // 边不存在，则添加


      if (!arc) {
        arc = new Arc(v1, v2, info);

        this.__addArc(v1, arc);

        this.__addArc(v2, arc);
      } // 对于已存在的边，更新边（权重等）数据
      else {
        arc.info = info;
      }
    }
    /**
     * @description 广度优先遍历
     *
     * @param {*} start 遍历的起始顶点
     * @param {function} callback 回调函数
     * @memberof UndirectedGraph
     */

  }, {
    key: "bfsTraverse",
    value: function bfsTraverse(start, callback) {
      if (this.vertexCount === 0) return;
      var vertex = null; // 参数检查：
      // --1. 两个参数的时候：指定起点和遍历的回调函数

      if (start && callback) {
        if (!(0, _toollib.isFunction)(callback)) throw new Error("".concat(callback, " is not a function."));
        vertex = this.__findVertex(start);
        if (vertex === undefined) throw new Error("vertex of ".concat(start, " doesnot exist."));
      } // --2.1 一个参数的时候，仅指定遍历的回调函数
      else if ((0, _toollib.isFunction)(start) && !callback) {
        callback = start;
        start = null;
        vertex = this.__vertexes[0];
      } // --2.2 两个参数的时候，仅指定遍历的回调函数
      else if (start == null && (0, _toollib.isFunction)(callback)) {
        vertex = this.__vertexes[0];
      } // --3. 参数错误的其它情况
      else {
        throw new Error("parameter error: need [vStart, callback] or [callback].\r\nreceived: ".concat(start, ", ").concat(callback));
      } // 广度优先：使用队列


      var queue = new _Queue.Queue(); // 初始化探索状态

      var map = this.__initVisitStatus();

      while (vertex) {
        // 如果当前顶点未被探索，则探索（回调访问）
        if (map.get(vertex) !== _GraphBase2.VisitStatus.VISITED) {
          var stop = callback(vertex.data);
          if (stop) return;
          map.set(vertex, _GraphBase2.VisitStatus.VISITED);
        }

        var arc = vertex.first;

        while (arc) {
          var otherVertex = arc.otherVertex(vertex); // 如果另一端顶点未被发现或探索，则加入队列并标记

          if (map.get(otherVertex) === _GraphBase2.VisitStatus.INIT) {
            queue.enqueue(otherVertex);
            map.set(otherVertex, _GraphBase2.VisitStatus.DISCOVERED);
          }

          arc = arc.nextArc(vertex);
        } // 下一层顶点


        vertex = queue.dequeue();
      }
    }
    /**
     * @description 深度优先遍历
     *
     * @param {*} start 遍历的起始顶点
     * @param {function} callback 回调函数
     * @memberof UndirectedGraph
     */

  }, {
    key: "dfsTraverse",
    value: function dfsTraverse(start, callback) {
      if (this.vertexCount === 0) return;
      var vertex = null; // 参数检查：
      // --1. 两个参数的时候：指定起点和遍历的回调函数

      if (start && callback) {
        if (!(0, _toollib.isFunction)(callback)) throw new Error("".concat(callback, " is not a function."));
        vertex = this.__findVertex(start);
        if (vertex === undefined) throw new Error("vertex of ".concat(start, " doesnot exist."));
      } // --2.1 一个参数的时候，仅指定遍历的回调函数
      else if ((0, _toollib.isFunction)(start) && !callback) {
        callback = start;
        start = null;
        vertex = this.__vertexes[0];
      } // --2.2 两个参数的时候，仅指定遍历的回调函数
      else if (start == null && (0, _toollib.isFunction)(callback)) {
        vertex = this.__vertexes[0];
      } // --3. 参数错误的其它情况
      else {
        throw new Error("parameter error: need [vStart, callback] or [callback].\r\nreceived: ".concat(start, ", ").concat(callback));
      } // 深度优先：使用栈


      var stack = new _Stack.Stack(); // 初始化探索状态

      var map = this.__initVisitStatus(); // 作为入口的首个顶点直接探索（回调）并标记


      var stop = callback(vertex.data);
      if (stop) return;
      map.set(vertex, _GraphBase2.VisitStatus.VISITED); // 回溯用

      stack.push(vertex);

      while (vertex) {
        var arc = vertex.first;

        while (arc) {
          // 根据深度优先原则，探索另一端顶点
          var otherVertex = arc.otherVertex(vertex); // 如果未探索，则探索（回调）并标记

          if (map.get(otherVertex) !== _GraphBase2.VisitStatus.VISITED) {
            stop = callback(otherVertex.data);
            if (stop) return;
            map.set(otherVertex, _GraphBase2.VisitStatus.VISITED); // 记录该顶点以备回溯

            vertex = otherVertex;
            arc = otherVertex.first;
            stack.push(vertex);
          } // 顶点已被发现或探索，寻找其它边
          else {
            arc = arc.nextArc(vertex);
          }
        } // 回溯


        vertex = stack.pop();
      }
    }
    /**
     * @description 查找边
     *
     * @param {VertexNode} v1
     * @param {VertexNode} v2
     * @returns Arc | null
     * @memberof UndirectedGraph
     */

  }, {
    key: "__findArc",
    value: function __findArc(v1, v2) {
      var arc = v1.first;

      while (arc) {
        if (arc.hasVertex(v2)) return arc;
        arc = arc.nextArc(v1);
      }

      return null;
    }
    /**
     * @description 内部函数，给顶点添加边
     *
     * @param {VertexNode} v 顶点
     * @param {Arc} arc 边
     * @memberof UndirectedGraph
     */

  }, {
    key: "__addArc",
    value: function __addArc(v, arc) {
      if (arc && !arc.hasVertex(v)) throw new Error("canot add arc:".concat(arc.toString(), " to vertex:").concat(v.toString()));
      var tempArc = v.first;
      var pre = tempArc;

      if (!tempArc) {
        v.first = arc;
      } else {
        while (tempArc) {
          pre = tempArc;
          tempArc = tempArc.nextArc(v);
        }

        if (pre.vertex1 === v) {
          pre.orderNext = arc;
        } else {
          pre.reverseNext = arc;
        }
      }
    }
  }]);

  return UndirectedGraph;
}(_GraphBase2.GraphBase);
/**
* @description 无向图内部用顶点类
* @class VertexNode
*/


exports.UndirectedGraph = UndirectedGraph;

var VertexNode = /*#__PURE__*/function () {
  function VertexNode(data) {
    _classCallCheck(this, VertexNode);

    this.data = data; // * data

    this.first = null; // Arc
  }

  _createClass(VertexNode, [{
    key: "toString",
    value: function toString() {
      return "".concat(this.data.toString());
    }
  }]);

  return VertexNode;
}();
/**
* @description 无向图内部用边类
* @class Arc
*/


var Arc = /*#__PURE__*/function () {
  function Arc(v1, v2, info) {
    _classCallCheck(this, Arc);

    this.vertex1 = v1; // VertexNode

    this.vertex2 = v2; // VertexNode

    this.orderNext = null; // Arc

    this.reverseNext = null; // Arc

    this.info = info; // *
  }
  /**
   * @description 指定顶点的下一条邻接边
   *
   * @param {VertexNode} v 顶点
   * @returns Arc | null
   * @memberof Arc
   */


  _createClass(Arc, [{
    key: "nextArc",
    value: function nextArc(v) {
      if (v === this.vertex1) {
        return this.orderNext;
      } else if (v === this.vertex2) {
        return this.reverseNext;
      }

      return null;
    }
    /**
     * @description 当前边是否包含指定顶点
     *
     * @param {VertexNode} v 顶点
     * @returns true | false
     * @memberof Arc
     */

  }, {
    key: "hasVertex",
    value: function hasVertex(v) {
      return v === this.vertex1 || v === this.vertex2;
    }
    /**
     * @description 指定顶点所在当前邻接边的另一端
     *
     * @param {VertexNode} v 顶点
     * @returns VertexNode | null
     * @memberof Arc
     */

  }, {
    key: "otherVertex",
    value: function otherVertex(v) {
      if (v === this.vertex1) {
        return this.vertex2;
      } else if (v === this.vertex2) {
        return this.vertex1;
      }

      return null;
    }
  }, {
    key: "toString",
    value: function toString() {
      return "\n          {\n              v1: ".concat(this.vertex1.toString(), ",\n              v2: ").concat(this.vertex2.toString(), ",\n              info: ").concat(this.info ? this.info.toString() : null, ",\n          }\n      ");
    }
  }]);

  return Arc;
}();