"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.needReversal = exports.loopToFind = exports.doublyFindAt = exports.LinkedListNode = void 0;

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); Object.defineProperty(Constructor, "prototype", { writable: false }); return Constructor; }

/*
 * @Author: coderdeh
 * @Date: 2022-07-11 15:55:50
 * @LastEditors: coderdeh
 * @LastEditTime: 2022-07-11 15:56:00
 * @FilePath: \coderdeh-vue\src\LinkedList\LinkedListNode.js
 * @Description: 
 */

/**
 * 链表数据结构内部用元素节点类
 * 注意：不应被 return 到外部
 *
 * @class Node
 */
var LinkedListNode = /*#__PURE__*/function () {
  function LinkedListNode(data) {
    var next = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
    var prev = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

    _classCallCheck(this, LinkedListNode);

    this.data = data;
    this.next = next;
    this.prev = prev;

    this.toString = function () {
      return this.data.toString.apply(this.data);
    };
  }

  _createClass(LinkedListNode, [{
    key: "toString",
    value: function toString() {
      return this.toString();
    }
  }]);

  return LinkedListNode;
}();
/**
* 链表内部用查找方法
*
* @param {number} position 位置下标
* @param {LinkedListNode} node 查找起始的元素节点
* @param {number} [lastIndex=0] 可选参数：链表末尾元素的下标
* @returns 指点位置的元素节点
*/


exports.LinkedListNode = LinkedListNode;

var loopToFind = function loopToFind(position, node) {
  var lastIndex = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
  var index = lastIndex || 0; // 根据参数[lastIndex]选择从头查找还是从尾查找

  if (lastIndex) {
    // seek from tail
    while (index-- > position) {
      node = node.prev;
    }
  } else {
    // seek from head
    while (index++ < position) {
      node = node.next;
    }
  }

  return node;
};
/**
* 检查指定的位置是否处于链表的后半部分
* 以改变查找的顺序
* @param {number} size 链表元素总个数
* @param {number} position 指定的位置
* @returns true:从末尾开始查找 false:从头部开始查找
*/


exports.loopToFind = loopToFind;

var needReversal = function needReversal(size, position) {
  var middle = Math.floor(size / 2);
  return middle < position;
};
/**
* 根据指定位置下标查找链表内元素
* 当指定位置处于链表后半部分时，
* 从尾部开始遍历查找，
* 否则从头部开始
* @param {number} size 链表容量
* @param {number} position 位置下标
* @param {LinkedListNode} head 首位元素
* @param {LinkedListNode} tail 末尾元素
* @returns 链表节点元素
*/


exports.needReversal = needReversal;

var doublyFindAt = function doublyFindAt(size, position, head, tail) {
  // 1.根据所在位置判断是否从末尾开始查找
  var reversal = needReversal(size, position); // 2.找到指定下标位置元素

  return reversal ? loopToFind(position, tail, size - 1) : loopToFind(position, head);
};

exports.doublyFindAt = doublyFindAt;